package exampler.fitlibrary;

import fitlibrary.parse.Table;

import java.util.Map;

public class BooleanFilterFixtureTests extends FilterFixtureTestCase {

    static class DummyFixture extends BooleanFilterFixture {

        protected DummyFixture() {
            this(new String[][] {});
        }

        protected DummyFixture(String[][] dataRows) {
            this("irrelevant fixture name", dataRows);
        }

        protected  DummyFixture(String derivedTableFixture, String[][] dataRows) {
            useTable(makeTable(derivedTableFixture, dataRows));
        }


        protected void setFlippersAndMaintainers() {
            setFlippersAndMaintainers(2, 1);  // 2nd and third columns
        }

        protected String columnDefault(int column) {
            return "default for " + String.valueOf(column);
        }

        protected String resultOfFlipping() {
            return "result-of-flipping";
        }
        protected String resultOfMaintaining() {
            return "result-of-maintaining";
        }

    }




    // The Incoming Table needs to be picked apart.

    public void testFindingDerivedFixtureName() {
        BooleanFilterFixture fixture = new DummyFixture("fit.ColumnFixtureDescendant", noRows());
        assertEquals("fit.ColumnFixtureDescendant",
                     fixture.derivedFixture());
    }

    public void testFindingColumnHeaders() {
        String[] expected = new String[]
                { "categoryA", "categoryB", "categoryC"};

        BooleanFilterFixture fixture = new DummyFixture(new String[][]{
                { "categoryA", "-", "-" },
                { "categoryB", "-", "-" },
                { "categoryC", "-", "-" } });

        String[] actual = fixture.derivedHeadings();
        assertEqualStringArrays(expected, actual);
    }

    public void testFindingMultivaluedColumnValues() {
        BooleanFilterFixture fixture = new DummyFixture(new String[][]{
                { "categoryA", "Amaint0,Amaint1",     "-" },
                { "categoryB", "Bmaint0",         "-" },
                { "categoryC", "Cmaint0,Cmaint1,Cmaint2", "-" } });
        Map/*<String,RepeatingStream>*/ actual = fixture.columnToMap(1);
        assertEqualStringArrays(new String[]{"Amaint0", "Amaint1"},
                                getStream(actual, "categoryA").baseArray());
        assertEqualStringArrays(new String[]{"Bmaint0"},
                                getStream(actual, "categoryB").baseArray());
        assertEqualStringArrays(new String[]{"Cmaint0", "Cmaint1", "Cmaint2"},
                                getStream(actual, "categoryC").baseArray());

    }

    // The DERIVED table is a 2D array that is later turned into a
    // fit Parse structure.

    // Part of the 2D array is the Flipper rows, which show
    // that a single value is able to flip the result of the
    // whole boolean expression.

    public void testNumberOfFlipperRowsIsGovernedByFlipperColumn() {
        BooleanFilterFixture fixture = new DummyFixture(new String[][]{
                { "categoryA", "Amaint0", "A-flip0" },
                { "categoryB", "Bmaint0", "B-flip0,B-flip1" },
                { "categoryC", "Cmaint0", "C-flip0" } });
        assertEquals(4, fixture.flipperRowCount());
    }

    public void testExcessMaintainersDoesNotChangeNumberOfFlipperRows() {
        BooleanFilterFixture fixture = new DummyFixture(new String[][]{
                { "categoryA", "Amaint0", "A-flip0" },
                { "categoryB", "Bmaint0", "B-flip0" },
                { "categoryC", "Cmaint0,Cmaint1", "B-flip0" } });
        assertEquals(3, fixture.flipperRowCount());
    }


    public void testCreationOfFlipperRowsFrom_SQUARE_TableWith_SINGLE_Values() {
        BooleanFilterFixture fixture = new DummyFixture(new String[][]{
                { "categoryA", "Amaint0", "A-flip0" },
                { "categoryB", "Bmaint0", "B-flip0" }});

        String[][] expected = new String[][] {
                { "A-flip0", "Bmaint0" },
                { "Amaint0", "B-flip0" } };

        String[][] actual = fixture.flipperRows();
        assertEqual2DStringArrays(expected, actual);
    }

    public void testCreationOfFlipperRowsFrom_SQUARE_TableWith_MULTIPLE_Values() {
        BooleanFilterFixture fixture = new DummyFixture(new String[][]{
                { "categoryA", "Amaint0,Amaint1", "A-flip0,A-flip1" },
                { "categoryB", "Bmaint0,Bmaint1", "B-flip0,B-flip1" }});

        String[][] expected = new String[][] { { "A-flip0", "Bmaint0" },
                                               { "A-flip1", "Bmaint1" },
                                               { "Amaint0", "B-flip0" },
                                               { "Amaint1", "B-flip1" }};

        String[][] actual = fixture.flipperRows();
        assertEqual2DStringArrays(expected, actual);
    }

    public void testCreationOfFlipperRowsFrom_EXCESS_COUNTEREXAMPLES_TableWith_MULTIPLE_Values() {
        BooleanFilterFixture fixture = new DummyFixture(new String[][]{
                { "categoryA", "Amaint0,Amaint1", "A-flip0,A-flip1,A-flip2" },
                { "categoryB", "Bmaint0,Bmaint1", "B-flip0,B-flip1" }});
        String[][] expected = new String[][] { { "A-flip0", "Bmaint0" },
                                               { "A-flip1", "Bmaint1" },
                                               { "A-flip2", "Bmaint0" },
                                               { "Amaint0", "B-flip0" },
                                               { "Amaint1", "B-flip1" }};

        String[][] actual = fixture.flipperRows();
        assertEqual2DStringArrays(expected, actual);
    }

    public void testCreationOfFlipperRowsFrom_EXCESS_EXAMPLES_TableWith_MULTIPLE_Values() {
        BooleanFilterFixture fixture = new DummyFixture(new String[][]{
                { "categoryA", "Amaint0,Amaint1,Amaint2", "A-flip0,A-flip1" },
                { "categoryB", "Bmaint0,Bmaint1,Bmaint2", "B-flip0,B-flip1,B-flip2" }});
        String[][] expected = new String[][] {
                { "A-flip0", "Bmaint0" },
                { "A-flip1", "Bmaint1" },
                { "Amaint0", "B-flip0" },
                { "Amaint1", "B-flip1" },
                { "Amaint2", "B-flip2" }};

        String[][] actual = fixture.flipperRows();
        assertEqual2DStringArrays(expected, actual);
    }

    public void testCreationOfFlipperRowsFromTableWith_MULTIPLE_ROWS() {
        BooleanFilterFixture fixture = new DummyFixture(new String[][]{
                { "categoryA", "Amaint0,Amaint1", "A-flip0" },
                { "categoryB", "Bmaint0", "B-flip0,B-flip1" },
                { "categoryC", "Cmaint0,Cmaint1,Cmaint2", "C-flip0, C-flip1" } });
        String[][] expected = new String[][] {
                { "A-flip0", "Bmaint0", "Cmaint0" },
                { "Amaint0", "B-flip0", "Cmaint1" },
                { "Amaint1", "B-flip1", "Cmaint2" },
                { "Amaint0", "Bmaint0", "C-flip0" },
                { "Amaint1", "Bmaint0", "C-flip1" }};

        String[][] actual = fixture.flipperRows();
        assertEqual2DStringArrays(expected, actual);
    }

    // Another part of the 2D array is the Maintainer rows, which show
    // that particular values do NOT flip the result.

    public void testMaintainerRowCountIsMaximumOfTheMaintainers() {
        BooleanFilterFixture fixture = new DummyFixture(new String[][]{
                { "categoryA", "Amaint0, Amaint1",      "A-flip0" },
                { "categoryB", "Bmaint0",           "B-flip0,B-flip1" },
                { "categoryC", "Cmaint0, Cmaint1, Cmaint2", "C-flip0, C-flip1, C-flip2, C-flip3, C-flip4, C-flip5" } });
        assertEquals(3, fixture.maintainerRowCount());
    }

    public void testCreationOfMaintainerRows() {
        BooleanFilterFixture fixture = new DummyFixture(new String[][]{
                { "categoryA", "Amaint0,Amaint1", "A-flip0" },
                { "categoryB", "Bmaint0", "B-flip0,B-flip1" },
                { "categoryC", "Cmaint0,Cmaint1,Cmaint2", "C-flip0, C-flip1" } });
        String[][] expected = new String[][] {
                { "Amaint0", "Bmaint0", "Cmaint0" },
                { "Amaint1", "Bmaint0", "Cmaint1" },
                { "Amaint0", "Bmaint0", "Cmaint2" } };

        String[][] actual = fixture.maintainerRows();
        assertEqual2DStringArrays(expected, actual);
    }

    public void testMaintainerRowsStartWithFirstMaintainerValue() {
        // Even though that's not the next unused maintainer value after
        // the last flipper row.
        BooleanFilterFixture fixture = new DummyFixture(new String[][]{
                { "categoryA", "Amaint0, Amaint1, Amaint2",   "A-flip0" },
                { "categoryB", "Bmaint0,Bmaint1",         "B-flip0,B-flip1" } });
        String[][] expected = new String[][] {
                { "Amaint0", "Bmaint0" },
                { "Amaint1", "Bmaint1" },
                { "Amaint2", "Bmaint0" } };

        fixture.flipperRows();  // this uses up some of the maintainer values
        String[][] actual = fixture.maintainerRows();
        assertEqual2DStringArrays(expected, actual);
    }

    // The derived table has to be turned into a true Fit Parse table.

    public void testResultTableIncludesAppropriateHeadings() {
        String[] columnHeadings = new String[] { "category A", "category B"};
        TableParse actual = (new DummyFixture()).assemble("derived fixture", columnHeadings, noRows(), noRows());
        Table indexable = new Table(actual);

        assertEquals(1, indexable.row(0).size());
        assertEquals("derived fixture", indexable.row(0).cell(0).text());
        assertEquals(3, indexable.row(1).size());
        assertEquals("category A", indexable.row(1).cell(0).text());
        assertEquals("category B", indexable.row(1).cell(1).text());
        assertEquals("result()", indexable.row(1).cell(2).text());
    }


    public void testResultTableIncludesTestRows() {
        String[] columnHeadings = new String[] { "category A", "category B"};
        String[][] flipperRows = new String[][] { {"Af0", "Bf0"},
                                                  {"Af1", "Bf1"} };
        String[][] maintainerRows = new String[][] { { "Am0", "Bm0" },
                                                     { "Am1", "Bm1" },
                                                     { "Am2", "Bm2" } };
        TableParse actual = (new DummyFixture()).assemble("derived fixture", columnHeadings, flipperRows, maintainerRows);

        assertExpectedTestRows(flipperRows, actual, RESULT_TABLE_NUM_HEADING_ROWS, "result-of-flipping");
        assertExpectedTestRows(maintainerRows, actual, RESULT_TABLE_NUM_HEADING_ROWS+2, "result-of-maintaining");
    }


    // Tables whose values are all true/false can omit the examples and counterexamples
    // columns

    public void testOmittedBooleanColumns() {
        BooleanFilterFixture fixture = new DummyFixture(new String[][]{
                { "defaulted" },
                { "given", "e1, e2", "c1,c2"} });

        Map/*<String,RepeatingStream>*/ actualOne = fixture.columnToMap(1);
        assertEqualStringArrays(new String[]{"default for 1"},
                                getStream(actualOne, "defaulted").baseArray());
        assertEqualStringArrays(new String[]{"e1", "e2"},
                                getStream(actualOne, "given").baseArray());

        Map/*<String,RepeatingStream>*/ actualTwo = fixture.columnToMap(2);
        assertEqualStringArrays(new String[]{"default for 2"},
                                getStream(actualTwo, "defaulted").baseArray());
        assertEqualStringArrays(new String[]{"c1", "c2"},
                                getStream(actualTwo, "given").baseArray());
    }




    private RepeatingStream getStream(Map actual, String category) {
        return ((RepeatingStream) actual.get(category));
    }
}